//**************************************************************
//*            OpenGLide - Glide to OpenGL Wrapper
//*             http://openglide.sourceforge.net
//*
//*                     Render File
//*
//*         OpenGLide is OpenSource under LGPL license
//*              Originaly made by Fabio Barros
//*      Modified by Paul for Glidos (http://www.glidos.net)
//*               Linux version by Simon White
//**************************************************************

#include "GlOgl.h"
#include "GLRender.h"
#include "PGTexture.h"

//**************************************************************
// Defines
//**************************************************************

#define DEBUG_MIN_MAX( var, maxvar, minvar )    \
    if ( var > maxvar ) maxvar = var;           \
    if ( var < minvar ) minvar = var;

//**************************************************************
// extern variables and functions prototypes
//**************************************************************

//**************************************************************
// Local variables
//**************************************************************

GLuint DisplayProgram;
GLuint DisplayVertexShader;
GLuint DisplayFragmentShader;

static const char *DisplayVertexShaderSource = \
	"#version 330\n"
	"\n"
	"layout(location = 0) in vec4 VPosition;\n"
	"\n"
	"void main()\n"
	"{\n"
	"	gl_Position = VPosition;\n"
	"}";

static const char *QuadBufferFragmentShaderSource = \
	"#version 330\n"
	"\n"
	"uniform sampler2DRect Sampler1;\n"
	"uniform sampler2DRect Sampler2;\n"
	"uniform int ScreenShift;\n"
	"layout(location = 0) out vec4 FragColor1;\n"
	"layout(location = 1) out vec4 FragColor2;\n"
	"\n"
	"void main()\n"
	"{\n"
	"	ivec2 p = ivec2(gl_FragCoord.xy);\n"
	"\n"
	"	FragColor1 = texelFetch(Sampler1, ivec2(p.x + ScreenShift, p.y));\n"
	"	FragColor2 = texelFetch(Sampler2, ivec2(p.x - ScreenShift, p.y));\n"
	"}";

static const char *HInterleavedFragmentShaderSource = \
	"#version 330\n"
	"\n"
	"uniform sampler2DRect Sampler1;\n"
	"uniform sampler2DRect Sampler2;\n"
	"uniform int ScreenShift;\n"
	"out vec4 FragColor;\n"
	"\n"
	"void main()\n"
	"{\n"
	"	ivec2 p = ivec2(gl_FragCoord.xy);\n"
	"\n"
	"	if ((p.y % 2) == 1)\n"
	"	{\n"
	"		p = p / 2;\n"
	"		p.x += ScreenShift;\n"
	"		p *= 2;\n"
	"		FragColor = (texelFetch(Sampler1, p) + texelFetch(Sampler1, ivec2(p.x + 1, p.y)) + texelFetch(Sampler1, ivec2(p.x, p.y + 1)) + texelFetch(Sampler1, p + 1)) / 4.0;\n"
	"	}\n"
	"	else\n"
	"	{\n"
	"		p = p / 2;\n"
	"		p.x -= ScreenShift;\n"
	"		p *= 2;\n"
	"		FragColor = (texelFetch(Sampler2, p) + texelFetch(Sampler2, ivec2(p.x + 1, p.y)) + texelFetch(Sampler2, ivec2(p.x, p.y + 1)) + texelFetch(Sampler2, p + 1)) / 4.0;\n"
	"	}\n"
	"}";

static const char *AnaglyphFragmentShaderSource = \
	"#version 330\n"
	"\n"
	"const mat4 ColorMatrix = mat4(\n"
	"	2.0 / 3.0, 1.0 / 6.0, 1.0 / 6.0, 0.0,\n"
	"	1.0 / 6.0, 2.0 / 3.0, 1.0 / 6.0, 0.0,\n"
	"	1.0 / 6.0, 1.0 / 6.0, 2.0 / 3.0, 0.0,\n"
	"	0.0, 0.0, 0.0, 1.0);\n"
	"uniform sampler2DRect Sampler1;\n"
	"uniform sampler2DRect Sampler2;\n"
	"uniform int ScreenShift;\n"
	"out vec4 FragColor;\n"
	"\n"
	"void main()\n"
	"{\n"
	"	ivec2 p = ivec2(gl_FragCoord.xy);\n"
	"\n"
	"	FragColor = vec4((ColorMatrix * texelFetch(Sampler1, ivec2(p.x + ScreenShift, p.y))).r, (ColorMatrix * texelFetch(Sampler2, ivec2(p.x - ScreenShift, p.y))).gb, 1.0);\n"
	"}";

GLuint RenderProgram;
GLuint RenderVertexShader;
GLuint RenderFragmentShader;

static const char *RenderVertexShaderSource = \
	"#version 330\n"
	"\n"
	"uniform mat4 TransformMatrix;\n"
	"layout(location = 0) in vec4 VPosition;\n"
	"layout(location = 1) in vec4 VColor1;\n"
	"layout(location = 2) in vec4 VColor2;\n"
	"layout(location = 3) in vec4 VTextureCoord;\n"
	"layout(location = 4) in float VFogCoord;\n"
	"out vec4 Color1;\n"
	"out vec4 Color2;\n"
	"out vec4 TextureCoord;\n"
	"out float FogCoord;\n"
	"\n"
	"void main()\n"
	"{\n"
	"	gl_Position = TransformMatrix * VPosition;\n"
	"	Color1 = VColor1;\n"
	"	Color2 = VColor2;\n"
	"	TextureCoord = VTextureCoord;\n"
	"	FogCoord = VFogCoord;\n"
	"}";

static const char *RenderFragmentShaderSource = \
	"#version 330\n"
	"\n"
	"in vec4 Color1;\n"
	"in vec4 Color2;\n"
	"uniform int TextureMode;\n"
	"uniform sampler2D Sampler;\n"
	"in vec4 TextureCoord;\n"
	"uniform int TextureEnvMode;\n"
	"uniform int FogMode;\n"
	"in float FogCoord;\n"
	"uniform vec4 FogColor;\n"
	"out vec4 FragColor;\n"
	"\n"
	"void main()\n"
	"{\n"
	"	vec4 c;\n"
	"	vec4 tc = textureProj(Sampler, TextureCoord);\n"
	"\n"
	"	if (TextureMode > 0)\n"
	"		switch (TextureEnvMode)\n"
	"		{\n"
	"			case 1:\n"
	"				c = mix(Color1, tc, tc.a) + Color2;\n"
	"				break;\n"
	"			case 2:\n"
	"				c = (tc + Color1) + Color2;\n"
	"				break;\n"
	"			case 3:\n"
	"				c = (tc * Color1) + Color2;\n"
	"				break;\n"
	"			case 4:\n"
	"				c = tc + Color2;\n"
	"				break;\n"
	"			default:\n"
	"				c = tc + Color2;\n"
	"				break;\n"
	"		}\n"
	"	else\n"
	"		c = Color1 + Color2;\n"
	"\n"
	"	if (FogMode == 1)\n"
	"		c = mix(c, FogColor, FogCoord);\n"
	"\n"
	"	FragColor = c;\n"
	"}";

GLint	TransformMatrix;

GLint	TextureMode;
GLint	TextureEnvMode;
GLint	FogMode;
GLint	FogColor;

GLuint	LeftFramebuffer, RightFramebuffer;
GLuint	LeftTexture, RightTexture;
GLuint	LeftDepth, RightDepth;

GLuint	DisplayVertexArray;
GLuint	DisplayArrayBuffer;

GLuint	RenderVertexArrays[256];
GLuint	RenderArrayBuffers[256];

GLuint	PixelUnpackBuffers[256];

GLfloat	LeftTransformMatrix[16];
GLfloat	RightTransformMatrix[16];

GLfloat ClipDepth;
GLfloat CullDepth;

// The functions for the color combine
ALPHAFACTORFUNCPROC AlphaFactorFunc;
COLORFACTORFUNCPROC ColorFactor3Func;
COLORFUNCTIONPROC   ColorFunctionFunc;

// Snapping constant
static const float vertex_snap_compare = 4096.0f;
static const float vertex_snap = float( 3L << 18 );

// Standard structs for the render
RenderStruct OGLRender;

// Varibles for the Add functions
static TColorStruct     Local, 
                        Other, 
                        CFactor;
static float            AFactor[3];
static float			*pVA;
static TVertexStruct	TVertex;
static TColorStruct		TColor1, TColor2;
static TTextureStruct	TTexture;
static TFogStruct		TFog;
static float            atmuoow;
static float            btmuoow;
static float            ctmuoow;
static float            aoow, 
                        boow, 
                        coow;
static float            hAspect, 
                        wAspect,
                        maxoow;

void OpenGLRenderInit();
void OpenGLRenderShutdown();
#if 0
void StereoShift(float Shift);
#endif

//**************************************************************
// Functions definitions
//**************************************************************

void OpenGLRenderInit()
{
	GLint i;
	GLchar s[256];
	GLfloat *bp;

	if (InternalConfig.StereoMode)
	{
		DisplayProgram = glCreateProgram();
		DisplayVertexShader = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(DisplayVertexShader, 1, &DisplayVertexShaderSource, NULL);
		glCompileShader(DisplayVertexShader);
		glGetShaderiv(DisplayVertexShader, GL_COMPILE_STATUS, &i);
		glGetShaderInfoLog(DisplayVertexShader, 256, NULL, s);
		GlideMsg("%s\n", s);
		if (!i)
		{
			Error("Error compiling vertex shader.\n");
			exit(1);
		}
		DisplayFragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
		switch (InternalConfig.StereoMode)
		{
			case 1:
				glShaderSource(DisplayFragmentShader, 1, &QuadBufferFragmentShaderSource, NULL);
				break;
			case 2:
			case 3:
				glShaderSource(DisplayFragmentShader, 1, &HInterleavedFragmentShaderSource, NULL);
				break;
			case 4:
				glShaderSource(DisplayFragmentShader, 1, &AnaglyphFragmentShaderSource, NULL);
				break;
		}
		glCompileShader(DisplayFragmentShader);
		glGetShaderiv(DisplayFragmentShader, GL_COMPILE_STATUS, &i);
		glGetShaderInfoLog(DisplayFragmentShader, 256, NULL, s);
		GlideMsg("%s\n", s);
		if (!i)
		{
			Error("Error compiling fragment shader.\n");
			exit(1);
		}
		glAttachShader(DisplayProgram, DisplayVertexShader);
		glAttachShader(DisplayProgram, DisplayFragmentShader);
		glLinkProgram(DisplayProgram);
		glGetProgramiv(DisplayProgram, GL_LINK_STATUS, &i);
		glGetProgramInfoLog(DisplayProgram, 256, NULL, s);
		GlideMsg("%s\n", s);
		if (!i)
		{
			Error("Error linking shader program.\n");
			exit(1);
		}

		glUseProgram(DisplayProgram);

		switch (UserConfig.StereoMode)
		{
			case 2:
				glUniform1i(glGetUniformLocation(DisplayProgram, "ScreenShift"), UserConfig.ScreenShift);
				glUniform1i(glGetUniformLocation(DisplayProgram, "Sampler1"), 0);
				glUniform1i(glGetUniformLocation(DisplayProgram, "Sampler2"), 1);
				glScissor(UserConfig.ScreenShift * 2, 0, OpenGL.WindowWidth - (UserConfig.ScreenShift * 2), OpenGL.WindowHeight);
				break;
			case 3:
				glUniform1i(glGetUniformLocation(DisplayProgram, "ScreenShift"), -UserConfig.ScreenShift);
				glUniform1i(glGetUniformLocation(DisplayProgram, "Sampler1"), 1);
				glUniform1i(glGetUniformLocation(DisplayProgram, "Sampler2"), 0);
				glScissor(UserConfig.ScreenShift * 2, 0, OpenGL.WindowWidth - (UserConfig.ScreenShift * 2), OpenGL.WindowHeight);
				break;
			default:
				glUniform1i(glGetUniformLocation(DisplayProgram, "ScreenShift"), UserConfig.ScreenShift);
				glUniform1i(glGetUniformLocation(DisplayProgram, "Sampler1"), 0);
				glUniform1i(glGetUniformLocation(DisplayProgram, "Sampler2"), 1);
				glScissor(UserConfig.ScreenShift, 0, OpenGL.WindowWidth - UserConfig.ScreenShift, OpenGL.WindowHeight);
				break;
		}
	}

	RenderProgram = glCreateProgram();
	RenderVertexShader = glCreateShader(GL_VERTEX_SHADER);
	glShaderSource(RenderVertexShader, 1, &RenderVertexShaderSource, NULL);
	glCompileShader(RenderVertexShader);
	glGetShaderiv(RenderVertexShader, GL_COMPILE_STATUS, &i);
	glGetShaderInfoLog(RenderVertexShader, 256, NULL, s);
	GlideMsg("%s\n", s);
	if (!i)
	{
		Error("Error compiling vertex shader.\n");
		exit(1);
	}
	RenderFragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
	glShaderSource(RenderFragmentShader, 1, &RenderFragmentShaderSource, NULL);
	glCompileShader(RenderFragmentShader);
	glGetShaderiv(RenderFragmentShader, GL_COMPILE_STATUS, &i);
	glGetShaderInfoLog(RenderFragmentShader, 256, NULL, s);
	GlideMsg("%s\n", s);
	if (!i)
	{
		Error("Error compiling fragment shader.\n");
		exit(1);
	}
	glAttachShader(RenderProgram, RenderVertexShader);
	glAttachShader(RenderProgram, RenderFragmentShader);
	glLinkProgram(RenderProgram);
	glGetProgramiv(RenderProgram, GL_LINK_STATUS, &i);
	glGetProgramInfoLog(RenderProgram, 256, NULL, s);
	GlideMsg("%s\n", s);
	if (!i)
	{
		Error("Error linking shader program.\n");
		exit(1);
	}

	glUseProgram(RenderProgram);

	TransformMatrix = glGetUniformLocation(RenderProgram, "TransformMatrix");
	TextureMode = glGetUniformLocation(RenderProgram, "TextureMode");
	TextureEnvMode = glGetUniformLocation(RenderProgram, "TextureEnvMode");
	FogMode = glGetUniformLocation(RenderProgram, "FogMode");
	FogColor = glGetUniformLocation(RenderProgram, "FogColor");

	if (InternalConfig.StereoMode)
	{
		glGenFramebuffers(1, &LeftFramebuffer);
		glBindFramebuffer(GL_FRAMEBUFFER, LeftFramebuffer);
		glGenTextures(1, &LeftTexture);
		glBindTexture(GL_TEXTURE_RECTANGLE, LeftTexture);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGBA, OpenGL.WindowWidth, OpenGL.WindowHeight, 0, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8_REV, NULL);
		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, LeftTexture, 0);
		glGenRenderbuffers(1, &LeftDepth);
		glBindRenderbuffer(GL_RENDERBUFFER, LeftDepth);
		glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, OpenGL.WindowWidth, OpenGL.WindowHeight);
		glBindRenderbuffer(GL_RENDERBUFFER, 0);
		glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, LeftDepth);
		if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
		{
			Error("Left framebuffer not complete");
			exit(1);
		}
		glDrawBuffer(GL_COLOR_ATTACHMENT0);

		glGenFramebuffers(1, &RightFramebuffer);
		glBindFramebuffer(GL_FRAMEBUFFER, RightFramebuffer);
		glGenTextures(1, &RightTexture);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_RECTANGLE, RightTexture);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGBA, OpenGL.WindowWidth, OpenGL.WindowHeight, 0, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8_REV, NULL);
		glActiveTexture(GL_TEXTURE0);
		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, RightTexture, 0);
		glGenRenderbuffers(1, &RightDepth);
		glBindRenderbuffer(GL_RENDERBUFFER, RightDepth);
		glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, OpenGL.WindowWidth, OpenGL.WindowHeight);
		glBindRenderbuffer(GL_RENDERBUFFER, 0);
		glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, RightDepth);
		if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
		{
			Error("Right framebuffer not complete");
			exit(1);
		}

		glGenVertexArrays(1, &DisplayVertexArray);
		glBindVertexArray(DisplayVertexArray);
		
		glGenBuffers(1, &DisplayArrayBuffer);
		glBindBuffer(GL_ARRAY_BUFFER, DisplayArrayBuffer);
		glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat) * 16, NULL, GL_STATIC_DRAW);
		do
		{
			bp = (GLfloat *)glMapBufferRange(GL_ARRAY_BUFFER, 0, sizeof(GLfloat) * 16, GL_MAP_WRITE_BIT | GL_MAP_UNSYNCHRONIZED_BIT);
			if (!bp)
			{
				Error("Failed to map buffer");
				exit(1);
			}
			*(bp++) = -1.0f;
			*(bp++) = 1.0f;
			*(bp++) = 0.0f;
			*(bp++) = 1.0f;
			*(bp++) = 1.0f;
			*(bp++) = 1.0f;
			*(bp++) = 0.0f;
			*(bp++) = 1.0f;
			*(bp++) = -1.0f;
			*(bp++) = -1.0f;
			*(bp++) = 0.0f;
			*(bp++) = 1.0f;
			*(bp++) = 1.0f;
			*(bp++) = -1.0f;
			*(bp++) = 0.0f;
			*bp = 1.0f;
		}
		while (!glUnmapBuffer(GL_ARRAY_BUFFER));

		glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, sizeof(GLfloat) * 4, (GLvoid *)0);
		glEnableVertexAttribArray(0);
	}

	memset(RenderVertexArrays, 0, sizeof(RenderVertexArrays));
	memset(RenderArrayBuffers, 0, sizeof(RenderArrayBuffers));

	memset(PixelUnpackBuffers, 0, sizeof(PixelUnpackBuffers));

	pVA = OGLRender.TVAttributes;
}

void OpenGLRenderShutdown()
{
	int i;

	if (InternalConfig.StereoMode)
	{
		glBindTexture(GL_TEXTURE_2D, 0);
		glBindRenderbuffer(GL_RENDERBUFFER, 0);
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
		glDeleteTextures(1, &LeftTexture);
		glDeleteRenderbuffers(1, &LeftDepth);
		glDeleteFramebuffers(1, &LeftFramebuffer);
		glDeleteTextures(1, &RightTexture);
		glDeleteRenderbuffers(1, &RightDepth);
		glDeleteFramebuffers(1, &RightFramebuffer);

		glBindVertexArray(0);
		glDeleteVertexArrays(1, &DisplayVertexArray);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
		for (i = 0; i < 256; i++)
		{
			if (RenderVertexArrays[i])
			{
				glDeleteVertexArrays(1, RenderVertexArrays + i);
				glDeleteBuffers(1, RenderArrayBuffers + i);
			}
			if (PixelUnpackBuffers[i])
				glDeleteBuffers(1, PixelUnpackBuffers + i);
		}

		glUseProgram(0);
		glDeleteShader(DisplayVertexShader);
		glDeleteShader(DisplayFragmentShader);
		glDeleteProgram(DisplayProgram);
		glDeleteShader(RenderVertexShader);
		glDeleteShader(RenderFragmentShader);
		glDeleteProgram(RenderProgram);
	}
	else
	{
		glBindVertexArray(0);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
		for (i = 0; i < 256; i++)
		{
			if (RenderVertexArrays[i])
			{
				glDeleteVertexArrays(1, RenderVertexArrays + i);
				glDeleteBuffers(1, RenderArrayBuffers + i);
			}
			if (PixelUnpackBuffers[i])
				glDeleteBuffers(1, PixelUnpackBuffers + i);
		}

		glUseProgram(0);
		glDeleteShader(RenderVertexShader);
		glDeleteShader(RenderFragmentShader);
		glDeleteProgram(RenderProgram);
	}
}

// Intializes the render and allocates memory
void RenderInitialize( void )
{

    OGLRender.NumberOfTriangles = 0;

#ifdef OGL_DEBUG
    OGLRender.FrameTriangles = 0;
    OGLRender.MaxTriangles = 0;
    OGLRender.MaxSequencedTriangles = 0;
    OGLRender.MinX = OGLRender.MinY = OGLRender.MinZ = OGLRender.MinW = 99999999.0f;
    OGLRender.MaxX = OGLRender.MaxY = OGLRender.MaxZ = OGLRender.MaxW = -99999999.0f;
    OGLRender.MinS = OGLRender.MinT = OGLRender.MinF = 99999999.0f;
    OGLRender.MaxS = OGLRender.MaxT = OGLRender.MaxF = -99999999.0f;

    OGLRender.MinR = OGLRender.MinG = OGLRender.MinB = OGLRender.MinA = 99999999.0f;
    OGLRender.MaxR = OGLRender.MaxG = OGLRender.MaxB = OGLRender.MaxA = -99999999.0f;
#endif
}

// Shutdowns the render and frees memory
void RenderFree( void )
{
}

void RenderDrawTriangles()
{
	int tdatasize;
	GLvoid *bdata;
    bool use_two_tex = false;

	if (!OGLRender.NumberOfTriangles)
		return;

	tdatasize = TRIANGLEDATASIZE * OGLRender.NumberOfTriangles;

	GetRenderVertexArray(tdatasize);

	do
	{
		bdata = glMapBufferRange(GL_ARRAY_BUFFER, 0, tdatasize, GL_MAP_WRITE_BIT | GL_MAP_UNSYNCHRONIZED_BIT | GL_MAP_INVALIDATE_BUFFER_BIT);
		if (!bdata)
		{
			Error("Failed to map buffer");
			exit(1);
		}
		memcpy(bdata, OGLRender.TVAttributes, tdatasize);
	}
	while (!glUnmapBuffer(GL_ARRAY_BUFFER));

    if ( OpenGL.Texture )
    {
        glUniform1i(TextureMode, GL_TRUE);

        use_two_tex = Textures->MakeReady( );

        if ( use_two_tex )
        {
			glActiveTexture( GL_TEXTURE1_ARB );

            glActiveTexture( GL_TEXTURE0_ARB );
        }
    }
    else
    {
        glUniform1i(TextureMode, GL_FALSE);;
    }

    if ( OpenGL.Blend )
    {
        glEnable( GL_BLEND );
    }
    else
    {
        glDisable( GL_BLEND );
    }

    // Alpha Fix
    if ( Glide.State.AlphaOther != GR_COMBINE_OTHER_TEXTURE )
    {
        glDisable( GL_ALPHA_TEST );
    }
    else 
    {
        if ( Glide.State.AlphaTestFunction != GR_CMP_ALWAYS )
        {
            glEnable( GL_ALPHA_TEST );
        }
    }
    
    if( !OpenGL.Blend && Glide.State.ChromaKeyMode )
    {
        glAlphaFunc( GL_GEQUAL, 0.5 );
        glEnable( GL_ALPHA_TEST );
        
		if (InternalConfig.StereoMode)
		{
			glBindFramebuffer(GL_DRAW_FRAMEBUFFER, LeftFramebuffer);
			glUniformMatrix4fv(TransformMatrix, 1, GL_FALSE, LeftTransformMatrix);
			glDrawArrays( GL_TRIANGLES, 0, OGLRender.NumberOfTriangles * 3 );

			glBindFramebuffer(GL_DRAW_FRAMEBUFFER, RightFramebuffer);
			glUniformMatrix4fv(TransformMatrix, 1, GL_FALSE, RightTransformMatrix);
			glDrawArrays( GL_TRIANGLES, 0, OGLRender.NumberOfTriangles * 3 );
		}
		else
			glDrawArrays( GL_TRIANGLES, 0, OGLRender.NumberOfTriangles * 3 );
        
        glDisable( GL_ALPHA_TEST );
    }
    else
		if (InternalConfig.StereoMode)
		{
			glBindFramebuffer(GL_DRAW_FRAMEBUFFER, LeftFramebuffer);
			glUniformMatrix4fv(TransformMatrix, 1, GL_FALSE, LeftTransformMatrix);
			glDrawArrays( GL_TRIANGLES, 0, OGLRender.NumberOfTriangles * 3 );

			glBindFramebuffer(GL_DRAW_FRAMEBUFFER, RightFramebuffer);
			glUniformMatrix4fv(TransformMatrix, 1, GL_FALSE, RightTransformMatrix);
			glDrawArrays( GL_TRIANGLES, 0, OGLRender.NumberOfTriangles * 3 );
		}
		else
			glDrawArrays( GL_TRIANGLES, 0, OGLRender.NumberOfTriangles * 3 );

    if ( use_two_tex )
    {
        glActiveTexture( GL_TEXTURE1_ARB );

        glActiveTexture( GL_TEXTURE0_ARB );
    }

#ifdef OGL_DEBUG
    if ( OGLRender.NumberOfTriangles > OGLRender.MaxSequencedTriangles )
    {
        OGLRender.MaxSequencedTriangles = OGLRender.NumberOfTriangles;
    }

    GLErro( "Render::DrawTriangles" );
#endif

	OGLRender.NumberOfTriangles = 0;
	pVA = OGLRender.TVAttributes;
}

void RenderAddTriangle( const GrVertex *a, const GrVertex *b, const GrVertex *c, bool unsnap )
{
    ZeroMemory( &TColor2, sizeof( TColorStruct ) );

	TVertex.aw = 1.0f;
	TVertex.bw = 1.0f;
	TVertex.cw = 1.0f;

    if ( ( Glide.State.STWHint & GR_STWHINT_W_DIFF_TMU0 ) == 0 )
    {
        atmuoow = a->oow;
        btmuoow = b->oow;
        ctmuoow = c->oow;
    }
    else
    {
        atmuoow = a->tmuvtx[ 0 ].oow;
        btmuoow = b->tmuvtx[ 0 ].oow;
        ctmuoow = c->tmuvtx[ 0 ].oow;
    }

    if ( Glide.ALocal )
    {
        switch ( Glide.State.AlphaLocal )
        {
        case GR_COMBINE_LOCAL_ITERATED:
            Local.aa = a->a * D1OVER255;
            Local.ba = b->a * D1OVER255;
            Local.ca = c->a * D1OVER255;
            break;

        case GR_COMBINE_LOCAL_CONSTANT:
            Local.aa = Local.ba = Local.ca = OpenGL.ConstantColor[ 3 ];
            break;

        case GR_COMBINE_LOCAL_DEPTH:
            Local.aa = a->z * D1OVER255;
            Local.ba = b->z * D1OVER255;
            Local.ca = c->z * D1OVER255;
            break;
        }
    }

    if ( Glide.AOther )
    {
        switch ( Glide.State.AlphaOther )
        {
        case GR_COMBINE_OTHER_ITERATED:
            Other.aa = a->a * D1OVER255;
            Other.ba = b->a * D1OVER255;
            Other.ca = c->a * D1OVER255;
            break;

        case GR_COMBINE_OTHER_CONSTANT:
            Other.aa = Other.ba = Other.ca = OpenGL.ConstantColor[ 3 ];
            break;

        case GR_COMBINE_OTHER_TEXTURE:
            if ( OpenGL.Texture )
            {
                Other.aa = Other.ba = Other.ca = 1.0f;
            }
            else
            {
                Other.aa = Other.ba = Other.ca = 0.0f;
            }
            break;
        }
    }

    if ( Glide.CLocal )
    {
        switch ( Glide.State.ColorCombineLocal )
        {
        case GR_COMBINE_LOCAL_ITERATED:
            Local.ar = a->r * D1OVER255;
            Local.ag = a->g * D1OVER255;
            Local.ab = a->b * D1OVER255;
            Local.br = b->r * D1OVER255;
            Local.bg = b->g * D1OVER255;
            Local.bb = b->b * D1OVER255;
            Local.cr = c->r * D1OVER255;
            Local.cg = c->g * D1OVER255;
            Local.cb = c->b * D1OVER255;
            break;

        case GR_COMBINE_LOCAL_CONSTANT:
            Local.ar = Local.br = Local.cr = OpenGL.ConstantColor[ 0 ];
            Local.ag = Local.bg = Local.cg = OpenGL.ConstantColor[ 1 ];
            Local.ab = Local.bb = Local.cb = OpenGL.ConstantColor[ 2 ];
            break;
        }
    }

    if ( Glide.COther )
    {
        switch ( Glide.State.ColorCombineOther )
        {
        case GR_COMBINE_OTHER_ITERATED:
            Other.ar = a->r * D1OVER255;
            Other.ag = a->g * D1OVER255;
            Other.ab = a->b * D1OVER255;
            Other.br = b->r * D1OVER255;
            Other.bg = b->g * D1OVER255;
            Other.bb = b->b * D1OVER255;
            Other.cr = c->r * D1OVER255;
            Other.cg = c->g * D1OVER255;
            Other.cb = c->b * D1OVER255;
            break;

        case GR_COMBINE_OTHER_CONSTANT:
            Other.ar = Other.br = Other.cr = OpenGL.ConstantColor[ 0 ];
            Other.ag = Other.bg = Other.cg = OpenGL.ConstantColor[ 1 ];
            Other.ab = Other.bb = Other.cb = OpenGL.ConstantColor[ 2 ];
            break;

        case GR_COMBINE_OTHER_TEXTURE:
            if ( OpenGL.Texture )
            {
                Other.ar = Other.ag = Other.ab = 1.0f;
                Other.br = Other.bg = Other.bb = 1.0f;
                Other.cr = Other.cg = Other.cb = 1.0f;
            }
            else
            {
                Other.ar = Other.ag = Other.ab = 0.0f;
                Other.br = Other.bg = Other.bb = 0.0f;
                Other.cr = Other.cg = Other.cb = 0.0f;
            }
            break;
        }
    }

    ColorFunctionFunc( &TColor1, &TColor2, &Local, &Other );

    switch ( Glide.State.AlphaFunction )
    {
    case GR_COMBINE_FUNCTION_ZERO:
        TColor1.aa = TColor1.ba = TColor1.ca = 0.0f;
        break;

    case GR_COMBINE_FUNCTION_LOCAL:
    case GR_COMBINE_FUNCTION_LOCAL_ALPHA:
        TColor1.aa = Local.aa;
        TColor1.ba = Local.ba;
        TColor1.ca = Local.ca;
        break;

    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        TColor1.aa = ( 1.0f - AlphaFactorFunc( Local.aa, Other.aa ) ) * Local.aa;
        TColor1.ba = ( 1.0f - AlphaFactorFunc( Local.ba, Other.ba ) ) * Local.ba;
        TColor1.ca = ( 1.0f - AlphaFactorFunc( Local.ca, Other.ca ) ) * Local.ca;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER:
        TColor1.aa = AlphaFactorFunc( Local.aa, Other.aa ) * Other.aa;
        TColor1.ba = AlphaFactorFunc( Local.ba, Other.ba ) * Other.ba;
        TColor1.ca = AlphaFactorFunc( Local.ca, Other.ca ) * Other.ca;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA:
        TColor1.aa = AlphaFactorFunc( Local.aa, Other.aa ) * Other.aa + Local.aa;
        TColor1.ba = AlphaFactorFunc( Local.ba, Other.ba ) * Other.ba + Local.ba;
        TColor1.ca = AlphaFactorFunc( Local.ca, Other.ca ) * Other.ca + Local.ca;
//      TColor2.aa =  Local.aa;
//      TColor2.ba =  Local.ba;
//      TColor2.ca =  Local.ca;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
        TColor1.aa = AlphaFactorFunc( Local.aa, Other.aa ) * ( Other.aa - Local.aa );
        TColor1.ba = AlphaFactorFunc( Local.ba, Other.ba ) * ( Other.ba - Local.ba );
        TColor1.ca = AlphaFactorFunc( Local.ca, Other.ca ) * ( Other.ca - Local.ca );
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        TColor1.aa = AlphaFactorFunc( Local.aa, Other.aa ) * ( Other.aa - Local.aa ) + Local.aa;
        TColor1.ba = AlphaFactorFunc( Local.ba, Other.ba ) * ( Other.ba - Local.ba ) + Local.ba;
        TColor1.ca = AlphaFactorFunc( Local.ca, Other.ca ) * ( Other.ca - Local.ca ) + Local.ca;
//      TColor2.aa =  Local.ba;
//      TColor2.ba =  Local.ba;
//      TColor2.ca =  Local.ca;
        break;
    }

    if ( Glide.State.ColorCombineInvert )
    {
        TColor1.ar = 1.0f - TColor1.ar - TColor2.ar;
        TColor1.ag = 1.0f - TColor1.ag - TColor2.ag;
        TColor1.ab = 1.0f - TColor1.ab - TColor2.ab;
        TColor1.br = 1.0f - TColor1.br - TColor2.br;
        TColor1.bg = 1.0f - TColor1.bg - TColor2.bg;
        TColor1.bb = 1.0f - TColor1.bb - TColor2.bb;
        TColor1.cr = 1.0f - TColor1.cr - TColor2.cr;
        TColor1.cg = 1.0f - TColor1.cg - TColor2.cg;
        TColor1.cb = 1.0f - TColor1.cb - TColor2.cb;
        TColor2.ar = TColor2.ag = TColor2.ab = 0.0f;
        TColor2.br = TColor2.bg = TColor2.bb = 0.0f;
        TColor2.cr = TColor2.cg = TColor2.cb = 0.0f;
    }

    if ( Glide.State.AlphaInvert )
    {
        TColor1.aa = 1.0f - TColor1.aa - TColor2.aa;
        TColor1.ba = 1.0f - TColor1.ba - TColor2.ba;
        TColor1.ca = 1.0f - TColor1.ca - TColor2.ca;
        TColor2.aa = TColor2.ba = TColor2.ca = 0.0f;
    }
    
    // Z-Buffering
    if ( ( Glide.State.DepthBufferMode == GR_DEPTHBUFFER_DISABLE ) || 
         ( Glide.State.DepthFunction == GR_CMP_ALWAYS ) )
    {
        TVertex.az = 0.0f;
        TVertex.bz = 0.0f;
        TVertex.cz = 0.0f;
    }
    else
    if ( OpenGL.DepthBufferType )
    {
        TVertex.az = a->ooz * D1OVER65535;
        TVertex.bz = b->ooz * D1OVER65535;
        TVertex.cz = c->ooz * D1OVER65535;
    }
    else
    {
       /*
        * For silly values of oow, depth buffering doesn't
        * seem to work, so map them to a sensible z.  When
        * games use these silly values, they probably don't
        * use z buffering anyway.
        */
        if ( a->oow > 1.0f )
        {
            TVertex.az = TVertex.bz = TVertex.cz = 1.0f;
        }
        else 
        if ( InternalConfig.PrecisionFix )
        {   // Fix precision to 16 integer bits.
            FxU16 w;
            w      = (FxU16)((a->oow / D1OVER65535) + 0.5f);
            TVertex.az = (float)w * D1OVER65535;
            w      = (FxU16)((b->oow / D1OVER65535) + 0.5f);
            TVertex.bz = (float)w * D1OVER65535;
            w      = (FxU16)((c->oow / D1OVER65535) + 0.5f);
            TVertex.cz = (float)w * D1OVER65535;
        }
        else
        {
            TVertex.az = a->oow;
            TVertex.bz = b->oow;
            TVertex.cz = c->oow;
        }
    }

	if (InternalConfig.StereoMode)
		if (ClipDepth > CullDepth)
		{
			if ((TVertex.az <= CullDepth) || (TVertex.bz <= CullDepth) || (TVertex.cz <= CullDepth))
			{
				TVertex.az = ClipDepth;
				TVertex.bz = ClipDepth;
				TVertex.cz = ClipDepth;
			}
		}
		else
		{
			if ((TVertex.az >= CullDepth) || (TVertex.bz >= CullDepth) || (TVertex.cz >= CullDepth))
			{
				TVertex.az = ClipDepth;
				TVertex.bz = ClipDepth;
				TVertex.cz = ClipDepth;
			}
		}

    if ( ( unsnap ) &&
         ( a->x > vertex_snap_compare ) )
    {
        TVertex.ax = a->x - vertex_snap;
        TVertex.ay = a->y - vertex_snap;
        TVertex.bx = b->x - vertex_snap;
        TVertex.by = b->y - vertex_snap;
        TVertex.cx = c->x - vertex_snap;
        TVertex.cy = c->y - vertex_snap;
    }
    else
    {
        TVertex.ax = a->x;
        TVertex.ay = a->y;
        TVertex.bx = b->x;
        TVertex.by = b->y;
        TVertex.cx = c->x;
        TVertex.cy = c->y;
    }

    if ( OpenGL.Texture )
    {
        maxoow = 1.0f / max( atmuoow, max( btmuoow, ctmuoow ) );

        Textures->GetAspect( &hAspect, &wAspect );

        TTexture.as = a->tmuvtx[ 0 ].sow * wAspect * maxoow;
        TTexture.at = a->tmuvtx[ 0 ].tow * hAspect * maxoow;
        TTexture.bs = b->tmuvtx[ 0 ].sow * wAspect * maxoow;
        TTexture.bt = b->tmuvtx[ 0 ].tow * hAspect * maxoow;
        TTexture.cs = c->tmuvtx[ 0 ].sow * wAspect * maxoow;
        TTexture.ct = c->tmuvtx[ 0 ].tow * hAspect * maxoow;

        TTexture.aq = TTexture.bq = TTexture.cq = 0.0f;
        TTexture.aoow = atmuoow * maxoow;
        TTexture.boow = btmuoow * maxoow;
        TTexture.coow = ctmuoow * maxoow;
    }

    if( InternalConfig.FogEnable )
    {
        if ( Glide.State.FogMode == GR_FOG_WITH_TABLE )
//        if ( Glide.State.FogMode & GR_FOG_WITH_TABLE )
        {
            TFog.af = (float)OpenGL.FogTable[ (FxU16)(1.0f / a->oow) ] * D1OVER255;
            TFog.bf = (float)OpenGL.FogTable[ (FxU16)(1.0f / b->oow) ] * D1OVER255;
            TFog.cf = (float)OpenGL.FogTable[ (FxU16)(1.0f / c->oow) ] * D1OVER255;
        }
        else
        {
            TFog.af = a->a * D1OVER255;
            TFog.bf = b->a * D1OVER255;
            TFog.cf = c->a * D1OVER255;
        }
//        if ( Glide.State.FogMode & GR_FOG_ADD2 )
//        {
//            TFog.af = 1.0f - TFog.af;
//            TFog.bf = 1.0f - TFog.bf;
//            TFog.cf = 1.0f - TFog.cf;
//        }
        
    #ifdef OGL_DEBUG
       DEBUG_MIN_MAX( TFog.af, OGLRender.MaxF, OGLRender.MinF );
       DEBUG_MIN_MAX( TFog.bf, OGLRender.MaxF, OGLRender.MinF );
       DEBUG_MIN_MAX( TFog.bf, OGLRender.MaxF, OGLRender.MinF );
    #endif
    }

#ifdef OGL_DEBUG
    DEBUG_MIN_MAX( TColor1.ar, OGLRender.MaxR, OGLRender.MinR );
    DEBUG_MIN_MAX( TColor1.br, OGLRender.MaxR, OGLRender.MinR );
    DEBUG_MIN_MAX( TColor1.cr, OGLRender.MaxR, OGLRender.MinR );
    
    DEBUG_MIN_MAX( TColor1.ag, OGLRender.MaxG, OGLRender.MinG );
    DEBUG_MIN_MAX( TColor1.bg, OGLRender.MaxG, OGLRender.MinG );
    DEBUG_MIN_MAX( TColor1.cg, OGLRender.MaxG, OGLRender.MinG );

    DEBUG_MIN_MAX( TColor1.ab, OGLRender.MaxB, OGLRender.MinB );
    DEBUG_MIN_MAX( TColor1.bb, OGLRender.MaxB, OGLRender.MinB );
    DEBUG_MIN_MAX( TColor1.cb, OGLRender.MaxB, OGLRender.MinB );

    DEBUG_MIN_MAX( TColor1.aa, OGLRender.MaxA, OGLRender.MinA );
    DEBUG_MIN_MAX( TColor1.ba, OGLRender.MaxA, OGLRender.MinA );
    DEBUG_MIN_MAX( TColor1.ca, OGLRender.MaxA, OGLRender.MinA );

    DEBUG_MIN_MAX( TVertex.az, OGLRender.MaxZ, OGLRender.MinZ );
    DEBUG_MIN_MAX( TVertex.bz, OGLRender.MaxZ, OGLRender.MinZ );
    DEBUG_MIN_MAX( TVertex.cz, OGLRender.MaxZ, OGLRender.MinZ );

    DEBUG_MIN_MAX( TVertex.ax, OGLRender.MaxX, OGLRender.MinX );
    DEBUG_MIN_MAX( TVertex.bx, OGLRender.MaxX, OGLRender.MinX );
    DEBUG_MIN_MAX( TVertex.cx, OGLRender.MaxX, OGLRender.MinX );

    DEBUG_MIN_MAX( TVertex.ay, OGLRender.MaxY, OGLRender.MinY );
    DEBUG_MIN_MAX( TVertex.by, OGLRender.MaxY, OGLRender.MinY );
    DEBUG_MIN_MAX( TVertex.cy, OGLRender.MaxY, OGLRender.MinY );

    DEBUG_MIN_MAX( TTexture.as, OGLRender.MaxS, OGLRender.MinS );
    DEBUG_MIN_MAX( TTexture.bs, OGLRender.MaxS, OGLRender.MinS );
    DEBUG_MIN_MAX( TTexture.cs, OGLRender.MaxS, OGLRender.MinS );

    DEBUG_MIN_MAX( TTexture.at, OGLRender.MaxT, OGLRender.MinT );
    DEBUG_MIN_MAX( TTexture.bt, OGLRender.MaxT, OGLRender.MinT );
    DEBUG_MIN_MAX( TTexture.ct, OGLRender.MaxT, OGLRender.MinT );

    OGLRender.FrameTriangles++;
#endif
    
    *(pVA++) = TVertex.ax;
	*(pVA++) = TVertex.ay;
	*(pVA++) = TVertex.az;
	*(pVA++) = TVertex.aw;
	*(pVA++) = TColor1.ar;
	*(pVA++) = TColor1.ag;
	*(pVA++) = TColor1.ab;
	*(pVA++) = TColor1.aa;
	*(pVA++) = TColor2.ar;
	*(pVA++) = TColor2.ag;
	*(pVA++) = TColor2.ab;
	*(pVA++) = TColor2.aa;
	*(pVA++) = TTexture.as;
	*(pVA++) = TTexture.at;
	*(pVA++) = TTexture.aq;
	*(pVA++) = TTexture.aoow;
	*(pVA++) = TFog.af;

	*(pVA++) = TVertex.bx;
	*(pVA++) = TVertex.by;
	*(pVA++) = TVertex.bz;
	*(pVA++) = TVertex.bw;
	*(pVA++) = TColor1.br;
	*(pVA++) = TColor1.bg;
	*(pVA++) = TColor1.bb;
	*(pVA++) = TColor1.ba;
	*(pVA++) = TColor2.br;
	*(pVA++) = TColor2.bg;
	*(pVA++) = TColor2.bb;
	*(pVA++) = TColor2.ba;
	*(pVA++) = TTexture.bs;
	*(pVA++) = TTexture.bt;
	*(pVA++) = TTexture.bq;
	*(pVA++) = TTexture.boow;
	*(pVA++) = TFog.bf;

	*(pVA++) = TVertex.cx;
	*(pVA++) = TVertex.cy;
	*(pVA++) = TVertex.cz;
	*(pVA++) = TVertex.cw;
	*(pVA++) = TColor1.cr;
	*(pVA++) = TColor1.cg;
	*(pVA++) = TColor1.cb;
	*(pVA++) = TColor1.ca;
	*(pVA++) = TColor2.cr;
	*(pVA++) = TColor2.cg;
	*(pVA++) = TColor2.cb;
	*(pVA++) = TColor2.ca;
	*(pVA++) = TTexture.cs;
	*(pVA++) = TTexture.ct;
	*(pVA++) = TTexture.cq;
	*(pVA++) = TTexture.coow;
	*(pVA++) = TFog.cf;

	OGLRender.NumberOfTriangles++;

    if ( OGLRender.NumberOfTriangles == MAXTRIANGLES )
		RenderDrawTriangles();

#ifdef OPENGL_DEBUG
    GLErro( "Render::AddTriangle" );
#endif
}

void RenderAddLine( const GrVertex *a, const GrVertex *b, bool unsnap )
{
#if 0
    // Color Stuff, need to optimize it
    ZeroMemory( &TColor2, sizeof( TColorStruct ) );

	TVertex.aw = 1.0f;
	TVertex.bw = 1.0f;
	TVertex.cw = 1.0f;

    if ( ( Glide.State.STWHint & GR_STWHINT_W_DIFF_TMU0 ) == 0 )
    {
        atmuoow = a->oow;
        btmuoow = b->oow;
    }
    else
    {
        atmuoow = a->tmuvtx[ 0 ].oow;
        btmuoow = b->tmuvtx[ 0 ].oow;
    }

    if ( Glide.ALocal )
    {
        switch ( Glide.State.AlphaLocal )
        {
        case GR_COMBINE_LOCAL_ITERATED:
            Local.aa = a->a * D1OVER255;
            Local.ba = b->a * D1OVER255;
            break;

        case GR_COMBINE_LOCAL_CONSTANT:
            Local.aa = Local.ba = OpenGL.ConstantColor[3];
            break;

        case GR_COMBINE_LOCAL_DEPTH:
            Local.aa = a->z;
            Local.ba = b->z;
            break;
        }
    }

    if ( Glide.AOther )
    {
        switch ( Glide.State.AlphaOther )
        {
        case GR_COMBINE_OTHER_ITERATED:
            Other.aa = a->a * D1OVER255;
            Other.ba = b->a * D1OVER255;
            break;

        case GR_COMBINE_OTHER_CONSTANT:
            Other.aa = Other.ba = OpenGL.ConstantColor[3];
            break;

        case GR_COMBINE_OTHER_TEXTURE:
            Other.aa = Other.ba = 1.0f;
            break;
        }
    }

    if ( Glide.CLocal )
    {
        switch ( Glide.State.ColorCombineLocal )
        {
        case GR_COMBINE_LOCAL_ITERATED:
            Local.ar = a->r * D1OVER255;
            Local.ag = a->g * D1OVER255;
            Local.ab = a->b * D1OVER255;
            Local.br = b->r * D1OVER255;
            Local.bg = b->g * D1OVER255;
            Local.bb = b->b * D1OVER255;
            break;

        case GR_COMBINE_LOCAL_CONSTANT:
            Local.ar = Local.br = OpenGL.ConstantColor[0];
            Local.ag = Local.bg = OpenGL.ConstantColor[1];
            Local.ab = Local.bb = OpenGL.ConstantColor[2];
            break;
        }
    }

    if ( Glide.COther )
    {
        switch ( Glide.State.ColorCombineOther )
        {
        case GR_COMBINE_OTHER_ITERATED:
            Other.ar = a->r * D1OVER255;
            Other.ag = a->g * D1OVER255;
            Other.ab = a->b * D1OVER255;
            Other.br = b->r * D1OVER255;
            Other.bg = b->g * D1OVER255;
            Other.bb = b->b * D1OVER255;
            break;

        case GR_COMBINE_OTHER_CONSTANT:
            Other.ar = Other.br = OpenGL.ConstantColor[0];
            Other.ag = Other.bg = OpenGL.ConstantColor[1];
            Other.ab = Other.bb = OpenGL.ConstantColor[2];
            break;

        case GR_COMBINE_OTHER_TEXTURE:
            Other.ar = Other.ag = Other.ab = 1.0f;
            Other.br = Other.bg = Other.bb = 1.0f;
            break;
        }
    }

    switch ( Glide.State.ColorCombineFunction )
    {
    case GR_COMBINE_FUNCTION_ZERO:
        TColor1.ar = TColor1.ag = TColor1.ab = 0.0f; 
        TColor1.br = TColor1.bg = TColor1.bb = 0.0f; 
        break;

    case GR_COMBINE_FUNCTION_LOCAL:
        TColor1.ar = Local.ar;
        TColor1.ag = Local.ag;
        TColor1.ab = Local.ab;
        TColor1.br = Local.br;
        TColor1.bg = Local.bg;
        TColor1.bb = Local.bb;
        break;

    case GR_COMBINE_FUNCTION_LOCAL_ALPHA:
        TColor1.ar = TColor1.ag = TColor1.ab = Local.aa;
        TColor1.br = TColor1.bg = TColor1.bb = Local.ba;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * Other.ar;
        TColor1.ag = CFactor.ag * Other.ag;
        TColor1.ab = CFactor.ab * Other.ab;
        TColor1.br = CFactor.br * Other.br;
        TColor1.bg = CFactor.bg * Other.bg;
        TColor1.bb = CFactor.bb * Other.bb;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * Other.ar;
        TColor1.ag = CFactor.ag * Other.ag;
        TColor1.ab = CFactor.ab * Other.ab;
        TColor1.br = CFactor.br * Other.br;
        TColor1.bg = CFactor.bg * Other.bg;
        TColor1.bb = CFactor.bb * Other.bb;
        TColor2.ar = Local.ar;
        TColor2.ag = Local.ag;
        TColor2.ab = Local.ab;
        TColor2.br = Local.br;
        TColor2.bg = Local.bg;
        TColor2.bb = Local.bb;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * Other.ar;
        TColor1.ag = CFactor.ag * Other.ag;
        TColor1.ab = CFactor.ab * Other.ab;
        TColor1.br = CFactor.br * Other.br;
        TColor1.bg = CFactor.bg * Other.bg;
        TColor1.bb = CFactor.bb * Other.bb;
        TColor2.ar = TColor2.ag = TColor2.ab = Local.aa;
        TColor2.br = TColor2.bg = TColor2.bb = Local.ba;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * (Other.ar - Local.ar);
        TColor1.ag = CFactor.ag * (Other.ag - Local.ag);
        TColor1.ab = CFactor.ab * (Other.ab - Local.ab);
        TColor1.br = CFactor.br * (Other.br - Local.br);
        TColor1.bg = CFactor.bg * (Other.bg - Local.bg);
        TColor1.bb = CFactor.bb * (Other.bb - Local.bb);
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
        if ((( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_ALPHA ) ||
            ( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_RGB )) &&
            (  Glide.State.ColorCombineOther == GR_COMBINE_OTHER_TEXTURE ) )
        {
            TColor1.ar = Local.ar;
            TColor1.ag = Local.ag;
            TColor1.ab = Local.ab;
            TColor1.br = Local.br;
            TColor1.bg = Local.bg;
            TColor1.bb = Local.bb;
        }
        else
        {
            ColorFactor3Func( &CFactor, &Local, &Other );
            TColor1.ar = CFactor.ar * (Other.ar - Local.ar);
            TColor1.ag = CFactor.ag * (Other.ag - Local.ag);
            TColor1.ab = CFactor.ab * (Other.ab - Local.ab);
            TColor1.br = CFactor.br * (Other.br - Local.br);
            TColor1.bg = CFactor.bg * (Other.bg - Local.bg);
            TColor1.bb = CFactor.bb * (Other.bb - Local.bb);
            TColor2.ar = Local.ar;
            TColor2.ag = Local.ag;
            TColor2.ab = Local.ab;
            TColor2.br = Local.br;
            TColor2.bg = Local.bg;
            TColor2.bb = Local.bb;
        }
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        if ((( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_ALPHA ) ||
            ( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_RGB )) &&
            (  Glide.State.ColorCombineOther == GR_COMBINE_OTHER_TEXTURE ) )
        {
            TColor1.ar = TColor1.ag = TColor1.ab = Local.aa;
            TColor1.br = TColor1.bg = TColor1.bb = Local.ba;
        }
        else
        {
            ColorFactor3Func( &CFactor, &Local, &Other );
            TColor1.ar = CFactor.ar * (Other.ar - Local.ar);
            TColor1.ag = CFactor.ag * (Other.ag - Local.ag);
            TColor1.ab = CFactor.ab * (Other.ab - Local.ab);
            TColor1.br = CFactor.br * (Other.br - Local.br);
            TColor1.bg = CFactor.bg * (Other.bg - Local.bg);
            TColor1.bb = CFactor.bb * (Other.bb - Local.bb);
            TColor2.ar = TColor2.ag = TColor2.ab = Local.aa;
            TColor2.br = TColor2.bg = TColor2.bb = Local.ba;
        }
        break;

    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = ( 1.0f - CFactor.ar ) * Local.ar;
        TColor1.ag = ( 1.0f - CFactor.ag ) * Local.ag;
        TColor1.ab = ( 1.0f - CFactor.ab ) * Local.ab;
        TColor1.br = ( 1.0f - CFactor.br ) * Local.br;
        TColor1.bg = ( 1.0f - CFactor.bg ) * Local.bg;
        TColor1.bb = ( 1.0f - CFactor.bb ) * Local.bb;
        TColor2.ar = Local.ar;
        TColor2.ag = Local.ag;
        TColor2.ab = Local.ab;
        TColor2.br = Local.br;
        TColor2.bg = Local.bg;
        TColor2.bb = Local.bb;
        break;

    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * -Local.ar;
        TColor1.ag = CFactor.ag * -Local.ag;
        TColor1.ab = CFactor.ab * -Local.ab;
        TColor1.br = CFactor.br * -Local.br;
        TColor1.bg = CFactor.bg * -Local.bg;
        TColor1.bb = CFactor.bb * -Local.bb;
        TColor2.ar = TColor2.ag = TColor2.ab = Local.aa;
        TColor2.br = TColor2.bg = TColor2.bb = Local.ba;
        break;
    }

    switch ( Glide.State.AlphaFunction )
    {
    case GR_COMBINE_FUNCTION_ZERO:
        TColor1.aa = TColor1.ba = 0.0f;
        break;

    case GR_COMBINE_FUNCTION_LOCAL:
    case GR_COMBINE_FUNCTION_LOCAL_ALPHA:
        TColor1.aa = Local.aa;
        TColor1.ba = Local.ba;
        break;

    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        TColor1.aa = ((1.0f - AlphaFactorFunc( Local.aa, Other.aa )) * Local.aa);
        TColor1.ba = ((1.0f - AlphaFactorFunc( Local.ba, Other.ba )) * Local.ba);
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER:
        TColor1.aa = (AlphaFactorFunc( Local.aa, Other.aa ) * Other.aa);
        TColor1.ba = (AlphaFactorFunc( Local.ba, Other.ba ) * Other.ba);
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA:
        TColor1.aa = (AlphaFactorFunc( Local.aa, Other.aa ) * Other.aa + Local.aa);
        TColor1.ba = (AlphaFactorFunc( Local.ba, Other.ba ) * Other.ba + Local.ba);
//      TColor2.aa =  Local.aa;
//      TColor2.ba =  Local.ba;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
        TColor1.aa = (AlphaFactorFunc( Local.aa, Other.aa ) * ( Other.aa - Local.aa ));
        TColor1.ba = (AlphaFactorFunc( Local.ba, Other.ba ) * ( Other.ba - Local.ba ));
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        TColor1.aa = (AlphaFactorFunc( Local.aa, Other.aa ) * ( Other.aa - Local.aa ) + Local.aa);
        TColor1.ba = (AlphaFactorFunc( Local.ba, Other.ba ) * ( Other.ba - Local.ba ) + Local.ba);
//      TColor2.aa =  Local.aa;
//      TColor2.ba =  Local.ba;
        break;
    }

    if ( Glide.State.ColorCombineInvert )
    {
        TColor1.ar = 1.0f - TColor1.ar - TColor2.ar;
        TColor1.ag = 1.0f - TColor1.ag - TColor2.ag;
        TColor1.ab = 1.0f - TColor1.ab - TColor2.ab;
        TColor1.br = 1.0f - TColor1.br - TColor2.br;
        TColor1.bg = 1.0f - TColor1.bg - TColor2.bg;
        TColor1.bb = 1.0f - TColor1.bb - TColor2.bb;
        TColor2.ar = TColor2.ag = TColor2.ab = 0.0f;
        TColor2.br = TColor2.bg = TColor2.bb = 0.0f;
    }

    if ( Glide.State.AlphaInvert )
    {
        TColor1.aa = 1.0f - TColor1.aa - TColor2.aa;
        TColor1.ba = 1.0f - TColor1.ba - TColor2.ba;
        TColor2.aa = TColor2.ba = 0.0f;
    }
    
    // Z-Buffering
    if ( ( Glide.State.DepthBufferMode == GR_DEPTHBUFFER_DISABLE ) || 
         ( Glide.State.DepthBufferMode == GR_CMP_ALWAYS ) )
    {
        TVertex.az = 0.0f;
        TVertex.bz = 0.0f;
    }
    else 
    if ( OpenGL.DepthBufferType )
    {
        TVertex.az = a->ooz * D1OVER65535;
        TVertex.bz = b->ooz * D1OVER65535;
    }
    else
    {
       /*
        * For silly values of oow, depth buffering doesn't
        * seem to work, so map them to a sensible z.  When
        * games use these silly values, they probably don't
        * use z buffering anyway.
        */
        if ( a->oow > 1.0f )
        {
            TVertex.az = TVertex.bz = 1.0f;
        }
        else 
        if ( InternalConfig.PrecisionFix )
        {   // Fix precision to 16 integer bits.
            FxU16 w;
            w      = (FxU16)((a->oow / D1OVER65535) + 0.5f);
            TVertex.az = (float)w * D1OVER65535;
            w      = (FxU16)((b->oow / D1OVER65535) + 0.5f);
            TVertex.bz = (float)w / D1OVER65535;
        }
        else
        {
            TVertex.az = a->oow;
            TVertex.bz = b->oow;
        }
    }

    if ( ( unsnap ) &&
         ( a->x > vertex_snap_compare ) )
    {
        TVertex.ax = a->x - vertex_snap;
        TVertex.ay = a->y - vertex_snap;
        TVertex.bx = b->x - vertex_snap;
        TVertex.by = b->y - vertex_snap;
    }
    else
    {
        TVertex.ax = a->x;
        TVertex.ay = a->y;
        TVertex.bx = b->x;
        TVertex.by = b->y;
    }

    if ( OpenGL.Texture )
    {
        Textures->GetAspect( &hAspect, &wAspect );

        TTexture.as = a->tmuvtx[0].sow * wAspect; // / a->oow;
        TTexture.at = a->tmuvtx[0].tow * hAspect; // / a->oow;
        TTexture.bs = b->tmuvtx[0].sow * wAspect; // / b->oow;
        TTexture.bt = b->tmuvtx[0].tow * hAspect; // / b->oow;

        TTexture.aq = TTexture.bq = 0.0f;
        TTexture.aoow = atmuoow;
        TTexture.boow = btmuoow;
    }

    if ( InternalConfig.FogEnable )
    {
        TFog.af = (float)OpenGL.FogTable[ (FxU16)(1.0f / a->oow) ] * D1OVER255;
        TFog.bf = (float)OpenGL.FogTable[ (FxU16)(1.0f / b->oow) ] * D1OVER255;

    #ifdef OGL_DEBUG
        DEBUG_MIN_MAX( TFog.af, OGLRender.MaxF, OGLRender.MinF );
        DEBUG_MIN_MAX( TFog.bf, OGLRender.MaxF, OGLRender.MinF );
    #endif
    }

#ifdef OGL_DEBUG
    DEBUG_MIN_MAX( TColor1.ar, OGLRender.MaxR, OGLRender.MinR );
    DEBUG_MIN_MAX( TColor1.br, OGLRender.MaxR, OGLRender.MinR );
    
    DEBUG_MIN_MAX( TColor1.ag, OGLRender.MaxG, OGLRender.MinG );
    DEBUG_MIN_MAX( TColor1.bg, OGLRender.MaxG, OGLRender.MinG );

    DEBUG_MIN_MAX( TColor1.ab, OGLRender.MaxB, OGLRender.MinB );
    DEBUG_MIN_MAX( TColor1.bb, OGLRender.MaxB, OGLRender.MinB );

    DEBUG_MIN_MAX( TColor1.aa, OGLRender.MaxA, OGLRender.MinA );
    DEBUG_MIN_MAX( TColor1.ba, OGLRender.MaxA, OGLRender.MinA );

    DEBUG_MIN_MAX( TVertex.az, OGLRender.MaxZ, OGLRender.MinZ );
    DEBUG_MIN_MAX( TVertex.bz, OGLRender.MaxZ, OGLRender.MinZ );

    DEBUG_MIN_MAX( TVertex.ax, OGLRender.MaxX, OGLRender.MinX );
    DEBUG_MIN_MAX( TVertex.bx, OGLRender.MaxX, OGLRender.MinX );

    DEBUG_MIN_MAX( TVertex.ay, OGLRender.MaxY, OGLRender.MinY );
    DEBUG_MIN_MAX( TVertex.by, OGLRender.MaxY, OGLRender.MinY );

    DEBUG_MIN_MAX( TTexture.as, OGLRender.MaxS, OGLRender.MinS );
    DEBUG_MIN_MAX( TTexture.bs, OGLRender.MaxS, OGLRender.MinS );

    DEBUG_MIN_MAX( TTexture.at, OGLRender.MaxT, OGLRender.MinT );
    DEBUG_MIN_MAX( TTexture.bt, OGLRender.MaxT, OGLRender.MinT );
#endif

    if ( OpenGL.Texture )
    {
        glUniform1i(TextureMode, GL_TRUE);

        Textures->MakeReady( );
    }
    else
    {
        glUniform1i(TextureMode, GL_FALSE);;
    }

    if ( OpenGL.Blend )
    {
        glEnable( GL_BLEND );
    }
    else
    {
        glDisable( GL_BLEND );
    }

    // Alpha Fix
    if ( Glide.State.AlphaOther != GR_COMBINE_OTHER_TEXTURE )
    {
        glDisable( GL_ALPHA_TEST );
    }
    else 
    {
        if ( Glide.State.AlphaTestFunction != GR_CMP_ALWAYS )
        {
            glEnable( GL_ALPHA_TEST );
        }
    }
    
    glBegin( GL_LINES );
        glColor4fv( &TColor1.ar );
        glSecondaryColor3fvEXT( &TColor2.ar );
        glTexCoord4fv( &TTexture.as );
        glFogCoordfEXT( TFog.af );
        glVertex3fv( &TVertex.ax );

        glColor4fv( &TColor1.br );
        glSecondaryColor3fvEXT( &TColor2.br );
        glTexCoord4fv( &TTexture.bs );
        glFogCoordfEXT( TFog.bf );
        glVertex3fv( &TVertex.bx );
    glEnd();

#ifdef OPENGL_DEBUG
    GLErro( "Render::AddLine" );
#endif
#endif
}

void RenderAddPoint( const GrVertex *a, bool unsnap )
{
#if 0
    // Color Stuff, need to optimize it
    ZeroMemory( &TColor2, sizeof( TColorStruct ) );

	TVertex.aw = 1.0f;
	TVertex.bw = 1.0f;
	TVertex.cw = 1.0f;

    if ( Glide.ALocal )
    {
        switch ( Glide.State.AlphaLocal )
        {
        case GR_COMBINE_LOCAL_ITERATED:
            Local.aa = a->a * D1OVER255;
            break;

        case GR_COMBINE_LOCAL_CONSTANT:
            Local.aa = OpenGL.ConstantColor[3];
            break;

        case GR_COMBINE_LOCAL_DEPTH:
            Local.aa = a->z;
            break;
        }
    }

    if ( Glide.AOther )
    {
        switch ( Glide.State.AlphaOther )
        {
        case GR_COMBINE_OTHER_ITERATED:
            Other.aa = a->a * D1OVER255;
            break;

        case GR_COMBINE_OTHER_CONSTANT:
            Other.aa = OpenGL.ConstantColor[3];
            break;

        case GR_COMBINE_OTHER_TEXTURE:
            Other.aa = 1.0f;
            break;
        }
    }

    if ( Glide.CLocal )
    {
        switch ( Glide.State.ColorCombineLocal )
        {
        case GR_COMBINE_LOCAL_ITERATED:
            Local.ar = a->r * D1OVER255;
            Local.ag = a->g * D1OVER255;
            Local.ab = a->b * D1OVER255;
            break;

        case GR_COMBINE_LOCAL_CONSTANT:
            Local.ar = OpenGL.ConstantColor[0];
            Local.ag = OpenGL.ConstantColor[1];
            Local.ab = OpenGL.ConstantColor[2];
            break;
        }
    }

    if ( Glide.COther )
    {
        switch ( Glide.State.ColorCombineOther )
        {
        case GR_COMBINE_OTHER_ITERATED:
            Other.ar = a->r * D1OVER255;
            Other.ag = a->g * D1OVER255;
            Other.ab = a->b * D1OVER255;
            break;

        case GR_COMBINE_OTHER_CONSTANT:
            Other.ar = OpenGL.ConstantColor[0];
            Other.ag = OpenGL.ConstantColor[1];
            Other.ab = OpenGL.ConstantColor[2];
            break;

        case GR_COMBINE_OTHER_TEXTURE:
            Other.ar = Other.ag = Other.ab = 1.0f;
            break;
        }
    }

    switch ( Glide.State.ColorCombineFunction )
    {
    case GR_COMBINE_FUNCTION_ZERO:
        TColor1.ar = TColor1.ag = TColor1.ab = 0.0f; 
        break;

    case GR_COMBINE_FUNCTION_LOCAL:
        TColor1.ar = Local.ar;
        TColor1.ag = Local.ag;
        TColor1.ab = Local.ab;
        break;

    case GR_COMBINE_FUNCTION_LOCAL_ALPHA:
        TColor1.ar = TColor1.ag = TColor1.ab = Local.aa;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * Other.ar;
        TColor1.ag = CFactor.ag * Other.ag;
        TColor1.ab = CFactor.ab * Other.ab;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * Other.ar;
        TColor1.ag = CFactor.ag * Other.ag;
        TColor1.ab = CFactor.ab * Other.ab;
        TColor2.ar = Local.ar;
        TColor2.ag = Local.ag;
        TColor2.ab = Local.ab;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * Other.ar;
        TColor1.ag = CFactor.ag * Other.ag;
        TColor1.ab = CFactor.ab * Other.ab;
        TColor2.ar = TColor2.ag = TColor2.ab = Local.aa;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * (Other.ar - Local.ar);
        TColor1.ag = CFactor.ag * (Other.ag - Local.ag);
        TColor1.ab = CFactor.ab * (Other.ab - Local.ab);
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
        if ((( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_ALPHA ) ||
            ( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_RGB )) &&
            (  Glide.State.ColorCombineOther == GR_COMBINE_OTHER_TEXTURE ) )
        {
            TColor1.ar = Local.ar;
            TColor1.ag = Local.ag;
            TColor1.ab = Local.ab;
        }
        else
        {
            ColorFactor3Func( &CFactor, &Local, &Other );
            TColor1.ar = CFactor.ar * (Other.ar - Local.ar);
            TColor1.ag = CFactor.ag * (Other.ag - Local.ag);
            TColor1.ab = CFactor.ab * (Other.ab - Local.ab);
            TColor2.ar = Local.ar;
            TColor2.ag = Local.ag;
            TColor2.ab = Local.ab;
        }
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        if ((( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_ALPHA ) ||
            ( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_RGB )) &&
            (  Glide.State.ColorCombineOther == GR_COMBINE_OTHER_TEXTURE ) )
        {
            TColor1.ar = TColor1.ag = TColor1.ab = Local.aa;
        }
        else
        {
            ColorFactor3Func( &CFactor, &Local, &Other );
            TColor1.ar = CFactor.ar * (Other.ar - Local.ar);
            TColor1.ag = CFactor.ag * (Other.ag - Local.ag);
            TColor1.ab = CFactor.ab * (Other.ab - Local.ab);
            TColor2.ar = TColor2.ag = TColor2.ab = Local.aa;
        }
        break;

    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = ( 1.0f - CFactor.ar ) * Local.ar;
        TColor1.ag = ( 1.0f - CFactor.ag ) * Local.ag;
        TColor1.ab = ( 1.0f - CFactor.ab ) * Local.ab;
        TColor2.ar = Local.ar;
        TColor2.ag = Local.ag;
        TColor2.ab = Local.ab;
        break;

    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        ColorFactor3Func( &CFactor, &Local, &Other );
        TColor1.ar = CFactor.ar * -Local.ar;
        TColor1.ag = CFactor.ag * -Local.ag;
        TColor1.ab = CFactor.ab * -Local.ab;
        TColor2.ar = TColor2.ag = TColor2.ab = Local.aa;
        break;
    }

    switch ( Glide.State.AlphaFunction )
    {
    case GR_COMBINE_FUNCTION_ZERO:
        TColor1.aa = 0.0f;
        break;

    case GR_COMBINE_FUNCTION_LOCAL:
    case GR_COMBINE_FUNCTION_LOCAL_ALPHA:
        TColor1.aa = Local.aa;
        break;

    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        TColor1.aa = ((1.0f - AlphaFactorFunc( Local.aa, Other.aa )) * Local.aa);
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER:
        TColor1.aa = (AlphaFactorFunc( Local.aa, Other.aa ) * Other.aa);
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA:
        TColor1.aa = AlphaFactorFunc( Local.aa, Other.aa ) * Other.aa + Local.aa;
//      TColor2.aa =  Local.aa;
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL:
        TColor1.aa = (AlphaFactorFunc( Local.aa, Other.aa ) * ( Other.aa - Local.aa ));
        break;

    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL:
    case GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA:
        TColor1.aa = AlphaFactorFunc( Local.aa, Other.aa ) * ( Other.aa - Local.aa ) + Local.aa;
//      TColor2.aa =  Local.aa;
        break;
    }

    if ( Glide.State.ColorCombineInvert )
    {
        TColor1.ar = 1.0f - TColor1.ar - TColor2.ar;
        TColor1.ag = 1.0f - TColor1.ag - TColor2.ag;
        TColor1.ab = 1.0f - TColor1.ab - TColor2.ab;
        TColor2.ar = TColor2.ag = TColor2.ab = 0.0f;
    }

    if ( Glide.State.AlphaInvert )
    {
        TColor1.aa = 1.0f - TColor1.aa - TColor2.aa;
        TColor2.aa = 0.0f;
    }
    
    // Z-Buffering
    if ( ( Glide.State.DepthBufferMode == GR_DEPTHBUFFER_DISABLE ) || 
         ( Glide.State.DepthBufferMode == GR_CMP_ALWAYS ) )
    {
        TVertex.az = 0.0f;
    }
    else 
    if ( OpenGL.DepthBufferType )
    {
        TVertex.az = a->ooz * D1OVER65535;
    }
    else
    {
        if ( a->oow > 1.0f )
        {
            TVertex.az = 1.0f;
        }
        else 
        if ( InternalConfig.PrecisionFix )
        {   // Fix precision to 16 integer bits.
            FxU16 w = (FxU16)((a->oow / D1OVER65535) + 0.5f);
            TVertex.az  = (float)w * D1OVER65535;
        }
        else
        {
            TVertex.az = a->oow;
        }
    }

    if ( ( unsnap ) &&
         ( a->x > vertex_snap_compare ) )
    {
        TVertex.ax = a->x - vertex_snap;
        TVertex.ay = a->y - vertex_snap;
    }
    else
    {
        TVertex.ax = a->x;
        TVertex.ay = a->y;
    }

    if ( OpenGL.Texture )
    {
        Textures->GetAspect( &hAspect, &wAspect );

        TTexture.as = a->tmuvtx[0].sow * wAspect;
        TTexture.at = a->tmuvtx[0].tow * hAspect;

        TTexture.aq = 0.0f;
        TTexture.aoow = a->oow;
    }

    if( InternalConfig.FogEnable )
    {
        TFog.af = (float)OpenGL.FogTable[ (FxU16)(1.0f / a->oow) ] * D1OVER255;

    #ifdef OGL_DEBUG
        DEBUG_MIN_MAX( TFog.af, OGLRender.MaxF, OGLRender.MinF );
    #endif
    }

#ifdef OGL_DEBUG
    DEBUG_MIN_MAX( TColor1.ar, OGLRender.MaxR, OGLRender.MinR );
    
    DEBUG_MIN_MAX( TColor1.ag, OGLRender.MaxG, OGLRender.MinG );

    DEBUG_MIN_MAX( TColor1.ab, OGLRender.MaxB, OGLRender.MinB );

    DEBUG_MIN_MAX( TColor1.aa, OGLRender.MaxA, OGLRender.MinA );

    DEBUG_MIN_MAX( TVertex.az, OGLRender.MaxZ, OGLRender.MinZ );

    DEBUG_MIN_MAX( TVertex.ax, OGLRender.MaxX, OGLRender.MinX );

    DEBUG_MIN_MAX( TVertex.ay, OGLRender.MaxY, OGLRender.MinY );

    DEBUG_MIN_MAX( TTexture.as, OGLRender.MaxS, OGLRender.MinS );

    DEBUG_MIN_MAX( TTexture.at, OGLRender.MaxT, OGLRender.MinT );
#endif

    if ( OpenGL.Texture )
    {
        glUniform1i(TextureMode, GL_TRUE);

        Textures->MakeReady( );
    }
    else
    {
        glUniform1i(TextureMode, GL_FALSE);;
    }

    if ( OpenGL.Blend )
    {
        glEnable( GL_BLEND );
    }
    else
    {
        glDisable( GL_BLEND );
    }

    // Alpha Fix
    if ( Glide.State.AlphaOther != GR_COMBINE_OTHER_TEXTURE )
    {
        glDisable( GL_ALPHA_TEST );
    }
    else 
    {
        if ( Glide.State.AlphaTestFunction != GR_CMP_ALWAYS )
        {
            glEnable( GL_ALPHA_TEST );
        }
    }
    
    glBegin( GL_POINTS );
        glColor4fv( &TColor1.ar );
        glSecondaryColor3fvEXT( &TColor2.ar );
        glTexCoord4fv( &TTexture.as );
        glFogCoordfEXT( TFog.af );
        glVertex3fv( &TVertex.ax );
    glEnd();

#ifdef OPENGL_DEBUG
    GLErro( "Render::AddPoint" );
#endif
#endif
}

// Color Factor functions

void FASTCALL ColorFactor3Zero( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = Result->ag = Result->ab = 0.0f;
    Result->br = Result->bg = Result->bb = 0.0f;
    Result->cr = Result->cg = Result->cb = 0.0f;
}

void FASTCALL ColorFactor3Local( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = ColorComponent->ar;
    Result->ag = ColorComponent->ag;
    Result->ab = ColorComponent->ab;
    Result->br = ColorComponent->br;
    Result->bg = ColorComponent->bg;
    Result->bb = ColorComponent->bb;
    Result->cr = ColorComponent->cr;
    Result->cg = ColorComponent->cg;
    Result->cb = ColorComponent->cb;
}

void FASTCALL ColorFactor3OtherAlpha( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = Result->ag = Result->ab = OtherAlpha->aa;
    Result->br = Result->bg = Result->bb = OtherAlpha->ba;
    Result->cr = Result->cg = Result->cb = OtherAlpha->ca;
}

void FASTCALL ColorFactor3LocalAlpha( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = Result->ag = Result->ab = ColorComponent->aa;
    Result->br = Result->bg = Result->bb = ColorComponent->ba;
    Result->cr = Result->cg = Result->cb = ColorComponent->ca;
}

void FASTCALL ColorFactor3OneMinusLocal( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = 1.0f - ColorComponent->ar;
    Result->ag = 1.0f - ColorComponent->ag;
    Result->ab = 1.0f - ColorComponent->ab;
    Result->br = 1.0f - ColorComponent->br;
    Result->bg = 1.0f - ColorComponent->bg;
    Result->bb = 1.0f - ColorComponent->bb;
    Result->cr = 1.0f - ColorComponent->cr;
    Result->cg = 1.0f - ColorComponent->cg;
    Result->cb = 1.0f - ColorComponent->cb;
}

void FASTCALL ColorFactor3OneMinusOtherAlpha( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = Result->ag = Result->ab = 1.0f - OtherAlpha->aa;
    Result->br = Result->bg = Result->bb = 1.0f - OtherAlpha->ba;
    Result->cr = Result->cg = Result->cb = 1.0f - OtherAlpha->ca;
}

void FASTCALL ColorFactor3OneMinusLocalAlpha( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = Result->ag = Result->ab = 1.0f - ColorComponent->aa;
    Result->br = Result->bg = Result->bb = 1.0f - ColorComponent->ba;
    Result->cr = Result->cg = Result->cb = 1.0f - ColorComponent->ca;
}

void FASTCALL ColorFactor3One( TColorStruct *Result, TColorStruct *ColorComponent, TColorStruct *OtherAlpha )
{
    Result->ar = Result->ag = Result->ab = 1.0f;
    Result->br = Result->bg = Result->bb = 1.0f;
    Result->cr = Result->cg = Result->cb = 1.0f;
}

// Alpha Factor functions

float AlphaFactorZero( float LocalAlpha, float OtherAlpha )
{
    return 0.0f;
}

float AlphaFactorLocal( float LocalAlpha, float OtherAlpha )
{
    return LocalAlpha;
}

float AlphaFactorOther( float LocalAlpha, float OtherAlpha )
{
    return OtherAlpha;
}

float AlphaFactorOneMinusLocal( float LocalAlpha, float OtherAlpha )
{
    return 1.0f - LocalAlpha;
}

float AlphaFactorOneMinusOther( float LocalAlpha, float OtherAlpha )
{
    return 1.0f - OtherAlpha;
}

float AlphaFactorOne( float LocalAlpha, float OtherAlpha )
{
    return 1.0f;
}

// Color functions

void ColorFunctionZero( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    pC->ar = pC->ag = pC->ab = 0.0f; 
    pC->br = pC->bg = pC->bb = 0.0f; 
    pC->cr = pC->cg = pC->cb = 0.0f; 
}

void ColorFunctionLocal( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    pC->ar = Local->ar;
    pC->ag = Local->ag;
    pC->ab = Local->ab;
    pC->br = Local->br;
    pC->bg = Local->bg;
    pC->bb = Local->bb;
    pC->cr = Local->cr;
    pC->cg = Local->cg;
    pC->cb = Local->cb;
}

void ColorFunctionLocalAlpha( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    pC->ar = pC->ag = pC->ab = Local->aa;
    pC->br = pC->bg = pC->bb = Local->ba;
    pC->cr = pC->cg = pC->cb = Local->ca;
}

void ColorFunctionScaleOther( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    ColorFactor3Func( &CFactor, Local, Other );
    pC->ar = CFactor.ar * Other->ar;
    pC->ag = CFactor.ag * Other->ag;
    pC->ab = CFactor.ab * Other->ab;
    pC->br = CFactor.br * Other->br;
    pC->bg = CFactor.bg * Other->bg;
    pC->bb = CFactor.bb * Other->bb;
    pC->cr = CFactor.cr * Other->cr;
    pC->cg = CFactor.cg * Other->cg;
    pC->cb = CFactor.cb * Other->cb;
}

void ColorFunctionScaleOtherAddLocal( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    ColorFactor3Func( &CFactor, Local, Other );
    pC->ar = CFactor.ar * Other->ar;
    pC->ag = CFactor.ag * Other->ag;
    pC->ab = CFactor.ab * Other->ab;
    pC->br = CFactor.br * Other->br;
    pC->bg = CFactor.bg * Other->bg;
    pC->bb = CFactor.bb * Other->bb;
    pC->cr = CFactor.cr * Other->cr;
    pC->cg = CFactor.cg * Other->cg;
    pC->cb = CFactor.cb * Other->cb;
    pC2->ar = Local->ar;
    pC2->ag = Local->ag;
    pC2->ab = Local->ab;
    pC2->br = Local->br;
    pC2->bg = Local->bg;
    pC2->bb = Local->bb;
    pC2->cr = Local->cr;
    pC2->cg = Local->cg;
    pC2->cb = Local->cb;
}

void ColorFunctionScaleOtherAddLocalAlpha( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    ColorFactor3Func( &CFactor, Local, Other );
    pC->ar = CFactor.ar * Other->ar;
    pC->ag = CFactor.ag * Other->ag;
    pC->ab = CFactor.ab * Other->ab;
    pC->br = CFactor.br * Other->br;
    pC->bg = CFactor.bg * Other->bg;
    pC->bb = CFactor.bb * Other->bb;
    pC->cr = CFactor.cr * Other->cr;
    pC->cg = CFactor.cg * Other->cg;
    pC->cb = CFactor.cb * Other->cb;
    pC2->ar = pC2->ag = pC2->ab = Local->aa;
    pC2->br = pC2->bg = pC2->bb = Local->ba;
    pC2->cr = pC2->cg = pC2->cb = Local->ca;
}

void ColorFunctionScaleOtherMinusLocal( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    ColorFactor3Func( &CFactor, Local, Other );
    pC->ar = CFactor.ar * (Other->ar - Local->ar);
    pC->ag = CFactor.ag * (Other->ag - Local->ag);
    pC->ab = CFactor.ab * (Other->ab - Local->ab);
    pC->br = CFactor.br * (Other->br - Local->br);
    pC->bg = CFactor.bg * (Other->bg - Local->bg);
    pC->bb = CFactor.bb * (Other->bb - Local->bb);
    pC->cr = CFactor.cr * (Other->cr - Local->cr);
    pC->cg = CFactor.cg * (Other->cg - Local->cg);
    pC->cb = CFactor.cb * (Other->cb - Local->cb);
}

void ColorFunctionScaleOtherMinusLocalAddLocal( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    if ((( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_ALPHA ) ||
        ( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_RGB )) &&
        (  Glide.State.ColorCombineOther == GR_COMBINE_OTHER_TEXTURE ) )
    {
        pC->ar = Local->ar;
        pC->ag = Local->ag;
        pC->ab = Local->ab;
        pC->br = Local->br;
        pC->bg = Local->bg;
        pC->bb = Local->bb;
        pC->cr = Local->cr;
        pC->cg = Local->cg;
        pC->cb = Local->cb;
    }
    else
    {
        ColorFactor3Func( &CFactor, Local, Other );
        pC->ar = CFactor.ar * (Other->ar - Local->ar);
        pC->ag = CFactor.ag * (Other->ag - Local->ag);
        pC->ab = CFactor.ab * (Other->ab - Local->ab);
        pC->br = CFactor.br * (Other->br - Local->br);
        pC->bg = CFactor.bg * (Other->bg - Local->bg);
        pC->bb = CFactor.bb * (Other->bb - Local->bb);
        pC->cr = CFactor.cr * (Other->cr - Local->cr);
        pC->cg = CFactor.cg * (Other->cg - Local->cg);
        pC->cb = CFactor.cb * (Other->cb - Local->cb);
        pC2->ar = Local->ar;
        pC2->ag = Local->ag;
        pC2->ab = Local->ab;
        pC2->br = Local->br;
        pC2->bg = Local->bg;
        pC2->bb = Local->bb;
        pC2->cr = Local->cr;
        pC2->cg = Local->cg;
        pC2->cb = Local->cb;
    }
}

void ColorFunctionScaleOtherMinusLocalAddLocalAlpha( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    if ((( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_ALPHA ) ||
        ( Glide.State.ColorCombineFactor == GR_COMBINE_FACTOR_TEXTURE_RGB )) &&
        (  Glide.State.ColorCombineOther == GR_COMBINE_OTHER_TEXTURE ) )
    {
        pC->ar = pC->ag = pC->ab = Local->aa;
        pC->br = pC->bg = pC->bb = Local->ba;
        pC->cr = pC->cg = pC->cb = Local->ca;
    }
    else
    {
        ColorFactor3Func( &CFactor, Local, Other );
        pC->ar = CFactor.ar * (Other->ar - Local->ar);
        pC->ag = CFactor.ag * (Other->ag - Local->ag);
        pC->ab = CFactor.ab * (Other->ab - Local->ab);
        pC->br = CFactor.br * (Other->br - Local->br);
        pC->bg = CFactor.bg * (Other->bg - Local->bg);
        pC->bb = CFactor.bb * (Other->bb - Local->bb);
        pC->cr = CFactor.cr * (Other->cr - Local->cr);
        pC->cg = CFactor.cg * (Other->cg - Local->cg);
        pC->cb = CFactor.cb * (Other->cb - Local->cb);
        pC2->ar = pC2->ag = pC2->ab = Local->aa;
        pC2->br = pC2->bg = pC2->bb = Local->ba;
        pC2->cr = pC2->cg = pC2->cb = Local->ca;
    }
}

void ColorFunctionMinusLocalAddLocal( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    ColorFactor3Func( &CFactor, Local, Other );
    pC->ar = ( 1.0f - CFactor.ar ) * Local->ar;
    pC->ag = ( 1.0f - CFactor.ag ) * Local->ag;
    pC->ab = ( 1.0f - CFactor.ab ) * Local->ab;
    pC->br = ( 1.0f - CFactor.br ) * Local->br;
    pC->bg = ( 1.0f - CFactor.bg ) * Local->bg;
    pC->bb = ( 1.0f - CFactor.bb ) * Local->bb;
    pC->cr = ( 1.0f - CFactor.cr ) * Local->cr;
    pC->cg = ( 1.0f - CFactor.cg ) * Local->cg;
    pC->cb = ( 1.0f - CFactor.cb ) * Local->cb;
    pC2->ar = Local->ar;
    pC2->ag = Local->ag;
    pC2->ab = Local->ab;
    pC2->br = Local->br;
    pC2->bg = Local->bg;
    pC2->bb = Local->bb;
    pC2->cr = Local->cr;
    pC2->cg = Local->cg;
    pC2->cb = Local->cb;
}

void ColorFunctionMinusLocalAddLocalAlpha( TColorStruct * pC, TColorStruct * pC2, TColorStruct * Local, TColorStruct * Other )
{
    ColorFactor3Func( &CFactor, Local, Other );
    pC->ar = CFactor.ar * -Local->ar;
    pC->ag = CFactor.ag * -Local->ag;
    pC->ab = CFactor.ab * -Local->ab;
    pC->br = CFactor.br * -Local->br;
    pC->bg = CFactor.bg * -Local->bg;
    pC->bb = CFactor.bb * -Local->bb;
    pC->cr = CFactor.cr * -Local->cr;
    pC->cg = CFactor.cg * -Local->cg;
    pC->cb = CFactor.cb * -Local->cb;
    pC2->ar = pC2->ag = pC2->ab = Local->aa;
    pC2->br = pC2->bg = pC2->bb = Local->ba;
    pC2->cr = pC2->cg = pC2->cb = Local->ca;
}

#if 0
void StereoShift(float Shift)
{
	TVertexStruct *p1, *p2, *p3;

	for (p1 = TVerticies, p2 = TVerticies + OGLRender.NumberOfTriangles, p3 = OGLRender.TVertex; p1 != p2; p1++, p3++)
	{
		if ((p3->az > 0.999) && (p3->bz > 0.999) && (p3->cz > 0.999))
		{
			p1->ax = p3->ax;
			p1->ay = p3->ay;
			p1->az = p3->az;
			p1->aw = 1;
			p1->bx = p3->bx;
			p1->by = p3->by;
			p1->bz = p3->bz;
			p1->bw = 1;
			p1->cx = p3->cx;
			p1->cy = p3->cy;
			p1->cz = p3->cz;
			p1->cw = 1;
		}
		else
		{
			p1->ax = p3->ax + (p3->az * Shift);
			p1->ay = p3->ay;
			p1->az = p3->az;
			p1->aw = 1;
			p1->bx = p3->bx + (p3->bz * Shift);
			p1->by = p3->by;
			p1->bz = p3->bz;
			p1->bw = 1;
			p1->cx = p3->cx + (p3->cz * Shift);
			p1->cy = p3->cy;
			p1->cz = p3->cz;
			p1->cw = 1;
		}
	}
}
#endif

void SetTransform(GLfloat Left, GLfloat Right, GLfloat Bottom, GLfloat Top, GLfloat Near, GLfloat Far)
{
	GLfloat m[16];

	if (InternalConfig.StereoMode)
	{
		ClipDepth = InternalConfig.CameraShift < 1.0f ? Far : Far + (Near - Far) / (InternalConfig.CameraShift / 32.0f);
		CullDepth = Near - ((Near - Far) / 256.0f);

		LeftTransformMatrix[0] = 2.0f / (Right - Left);
		RightTransformMatrix[0] = 2.0f / (Right - Left);
		LeftTransformMatrix[1] = 0.0f;
		RightTransformMatrix[1] = 0.0f;
		LeftTransformMatrix[2] = 0.0f;
		RightTransformMatrix[2] = 0.0f;
		LeftTransformMatrix[3] = 0.0f;
		RightTransformMatrix[3] = 0.0f;
		LeftTransformMatrix[4] = 0.0f;
		RightTransformMatrix[4] = 0.0f;
		LeftTransformMatrix[5] = 2.0f / (Top - Bottom);
		RightTransformMatrix[5] = 2.0f / (Top - Bottom);
		LeftTransformMatrix[6] = 0.0f;
		RightTransformMatrix[6] = 0.0f;
		LeftTransformMatrix[7] = 0.0f;
		RightTransformMatrix[7] = 0.0f;
		LeftTransformMatrix[8] = (Near - Far) * (InternalConfig.CameraShift * (2.0f / (Right - Left)));
		RightTransformMatrix[8] = (Far - Near) * (InternalConfig.CameraShift * (2.0f / (Right - Left)));
		LeftTransformMatrix[9] = 0.0f;
		RightTransformMatrix[9] = 0.0f;
		LeftTransformMatrix[10] = 2.0f / (Far - Near);
		RightTransformMatrix[10] = 2.0f / (Far - Near);
		LeftTransformMatrix[11] = 0.0f;
		RightTransformMatrix[11] = 0.0f;
		LeftTransformMatrix[12] = (Left + Right) / (Left - Right);
		RightTransformMatrix[12] = (Left + Right) / (Left - Right);
		LeftTransformMatrix[13] = (Bottom + Top) / (Bottom - Top);
		RightTransformMatrix[13] = (Bottom + Top) / (Bottom - Top);
		LeftTransformMatrix[14] = (Near + Far) / (Near - Far);
		RightTransformMatrix[14] = (Near + Far) / (Near - Far);
		LeftTransformMatrix[15] = 1.0f;
		RightTransformMatrix[15] = 1.0f;
	}
	else
	{
		m[0] = 2.0f / (Right - Left);
		m[1] = 0.0f;
		m[2] = 0.0f;
		m[3] = 0.0f;
		m[4] = 0.0f;
		m[5] = 2.0f / (Top - Bottom);
		m[6] = 0.0f;
		m[7] = 0.0f;
		m[8] = 0.0f;
		m[9] = 0.0f;
		m[10] = 2.0f / (Far - Near);
		m[11] = 0.0f;
		m[12] = (Left + Right) / (Left - Right);
		m[13] = (Bottom + Top) / (Bottom - Top);
		m[14] = (Near + Far) / (Near - Far);
		m[15] = 1.0f;
	}

	glUniformMatrix4fv(TransformMatrix, 1, GL_FALSE, m);
}

void GetRenderVertexArray(int Size)
{
	int bsize;
	int index;

	bsize = 1;
	index = 0;
	while (bsize < Size)
	{
		index++;
		if (index == 256)
		{
			Error("Render array buffer overflow");
			exit(1);
		}
		bsize *= 2;
	}

	if (RenderVertexArrays[index])
	{
		glBindVertexArray(RenderVertexArrays[index]);
		glBindBuffer(GL_ARRAY_BUFFER, RenderArrayBuffers[index]);
	}
	else
	{
		glGenVertexArrays(1, RenderVertexArrays + index);
		glBindVertexArray(RenderVertexArrays[index]);

		glGenBuffers(1, RenderArrayBuffers + index);
		glBindBuffer(GL_ARRAY_BUFFER, RenderArrayBuffers[index]);
		glBufferData(GL_ARRAY_BUFFER, bsize, NULL, GL_STREAM_DRAW);

		glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, VERTEXDATASIZE, (GLvoid *)0);
		glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, VERTEXDATASIZE, (GLvoid *)(sizeof(GLfloat) * 4));
		glVertexAttribPointer(2, 4, GL_FLOAT, GL_FALSE, VERTEXDATASIZE, (GLvoid *)(sizeof(GLfloat) * 8));
		glVertexAttribPointer(3, 4, GL_FLOAT, GL_FALSE, VERTEXDATASIZE, (GLvoid *)(sizeof(GLfloat) * 12));
		glVertexAttribPointer(4, 1, GL_FLOAT, GL_FALSE, VERTEXDATASIZE, (GLvoid *)(sizeof(GLfloat) * 16));
		glEnableVertexAttribArray(0);
		glEnableVertexAttribArray(1);
		glEnableVertexAttribArray(2);
		glEnableVertexAttribArray(3);
		glEnableVertexAttribArray(4);
	}
}

void GetPixelUnpackBuffer(int Size)
{
	int bsize;
	int index;

	bsize = 1;
	index = 0;
	while (bsize < Size)
	{
		index++;
		if (index == 256)
		{
			Error("Pixel unpack buffer overflow");
			exit(1);
		}
		bsize *= 2;
	}

	if (PixelUnpackBuffers[index])
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, PixelUnpackBuffers[index]);
	else
	{
		glGenBuffers(1, PixelUnpackBuffers + index);
		glBindBuffer(GL_PIXEL_UNPACK_BUFFER, PixelUnpackBuffers[index]);
		glBufferData(GL_PIXEL_UNPACK_BUFFER, bsize, NULL, GL_STREAM_DRAW);
	}
}